#ifndef __Xi_TabList_h__
#define __Xi_TabList_h__

#if defined(VMS) || defined(__VMS)
#include <X11/apienvset.h>
#endif

typedef enum {XiPIXMAP_TOP, XiPIXMAP_BOTTOM, XiPIXMAP_RIGHT, XiPIXMAP_LEFT,
	      XiPIXMAP_NONE, XiPIXMAP_ONLY} XiPixmapPlacement;
typedef enum {XiTAB_VALUE_COPY, XiTAB_VALUE_SHARE} XiTabValueMode;

typedef enum {XiTAB_CMP_VISUAL, XiTAB_CMP_SIZE, XiTAB_CMP_EQUAL} XiTabResult;

#define XiCOLOR_DYNAMIC ((Pixel)-1)
#define XiPIXMAP_DYNAMIC ((Pixmap) 3)
#define XiTAB_LAST_POSITION -1
#define XiTAB_NOT_FOUND -1

#define XiTAB_LABEL_STRING      (1L<<0)
#define XiTAB_LABEL_PIXMAP      (1L<<1)
#define XiTAB_PIXMAP_PLACEMENT  (1L<<2)
#define XiTAB_BACKGROUND        (1L<<3)
#define XiTAB_FOREGROUND        (1L<<4)
#define XiTAB_VALUE_MODE        (1L<<5)
#define XiTAB_LABEL_ALIGNMENT   (1L<<6)
#define XiTAB_STRING_DIRECTION  (1L<<7)
#define XiTAB_BACKGROUND_PIXMAP	(1L<<8)
#define XiTAB_SENSITIVE         (1L<<9)
#define XiTAB_ALL_FLAGS         (XiTAB_LABEL_STRING|XiTAB_LABEL_PIXMAP|\
				 XiTAB_PIXMAP_PLACEMENT|XiTAB_BACKGROUND|\
				 XiTAB_FOREGROUND|XiTAB_VALUE_MODE|\
				 XiTAB_LABEL_ALIGNMENT|XiTAB_STRING_DIRECTION|\
				 XiTAB_BACKGROUND_PIXMAP|XiTAB_SENSITIVE)

typedef struct _XiTabAttributeRec {
    XmString          label_string;	/* default: NULL                     */
    XmStringDirection string_direction; /* default: XmSTRING_DIRECTION_L_TO_R*/
    Pixmap            label_pixmap;	/* default: XmUNSPECIFIED_PIXMAP     */
    int               label_alignment;  /* default: XmALIGNEMENT_CENTER      */
    XiPixmapPlacement pixmap_placement; /* default: XiPIXMAP_RIGHT           */
    Pixel             foreground;       /* default: XiCOLOR_DYNAMIC          */
    Pixel             background;       /* default: XiCOLOR_DYNAMIC          */
    Pixmap            background_pixmap;/* default: XiPIXMAP_DYNAMIC         */
    Boolean	      sensitive;        /* default: True                     */
    XiTabValueMode    value_mode;       /* default: XiTAB_VALUE_COPY         */
} XiTabAttributeRec, * XiTabAttributes;

typedef struct _XiTabListRec *XiTabList;

XiTabList XiTabListCreate(
#ifndef _NO_PROTO
void
#endif
);

XiTabList XiTabListCopy(
#ifndef _NO_PROTO
XiTabList
#endif
);

void XiTabListFree(
#ifndef _NO_PROTO
XiTabList
#endif
);

void XiTabListRemove(
#ifndef _NO_PROTO
XiTabList, int
#endif
);

int XiTabListInsert(
#ifndef _NO_PROTO
XiTabList, int, XtValueMask, XiTabAttributes
#endif
);

int XiTabListAppend(
#ifndef _NO_PROTO
XiTabList, XtValueMask, XiTabAttributes
#endif
);

void XiTabListModify(
#ifndef _NO_PROTO
XiTabList, int, XtValueMask, XiTabAttributes
#endif
);

void XiTabListQuery(
#ifndef _NO_PROTO
XiTabList, int, XiTabAttributes
#endif
);

int XiTabListFind(
#ifndef _NO_PROTO
XiTabList, XmString
#endif
);

void XiTabListSimpleRemove(
#ifndef _NO_PROTO
XiTabList, XmString
#endif
);

int XiTabListSimpleInsert(
#ifndef _NO_PROTO
XiTabList, int, XmString
#endif
);

int XiTabListSimpleAppend(
#ifndef _NO_PROTO
XiTabList, XmString
#endif
);

void XiTabListSimpleModify(
#ifndef _NO_PROTO
XiTabList, int, XmString
#endif
);

XmString XiTabListSimpleQuery(
#ifndef _NO_PROTO
XiTabList, int
#endif
);

XiTabResult XiTabListCompare(
#ifndef _NO_PROTO
XiTabList, XiTabList
#endif
);

void XiTabAttibutesFree(
#ifndef _NO_PROTO
XiTabAttributes
#endif
);

#if defined(VMS) || defined(__VMS)
#include <X11/apienvrst.h>
#endif

#endif /* __Xi_TabList_h__ */
