/*
 *    Copyright 1990, Integrated Computer Solutions, Inc.
 *
 *		       All Rights Reserved.
 *
 * AUTHOR: Chris D. Peterson
 *
 */

#ifndef __XiP_h__
#define __XiP_h__

#if defined(VMS) || defined(__VMS)
#include <X11/apienvset.h>
#endif

#include <Xi/Xi.h>

/************************************************************
*	INCLUDE FILES
*************************************************************/

#if defined(hpux) && OS_MAJOR_VERSION < 10
#include <nl_ctype.h>
#endif

/************************************************************
*	TYPEDEFS AND DEFINES
*************************************************************/

typedef void (*XiVoidFunc)();
typedef int (*XiIntFunc)();
typedef unsigned int (*XiUnsignedIntFunc)();
typedef Widget (*XiWidgetFunc)();

extern String xi_std_filter[], xi_std_constraint_filter[];

/************************************************************
*	MACROS
*************************************************************/

#define streq(a, b) (((a) != NULL) && ((b) != NULL) && (strcmp((a), (b)) == 0))

#define ForAllChildren(w, childP)                                          \
  for ( (childP) = (w)->composite.children ;                               \
        (childP) < (w)->composite.children + (w)->composite.num_children ; \
        (childP)++ )

/*
 * Math Stuff 
 *
 * Some Systems define MIN and MAX so I have to undef them before I make
 * my own definitions.
 */

#undef MIN			
#undef MAX			
#undef ABS

#define MIN(a,b)        (((int)(a) < (int)(b)) ? (a) : (b))
#define MAX(a,b)        (((int)(a) > (int)(b)) ? (a) : (b))
#define ABS(a)          (((int)(a) >= 0) ? (a) : -(a))

#define ASSIGN_MAX(a, b) ((a) = ((int)(a) > (int)(b) ? (a) : (b)))
#define ASSIGN_MIN(a, b) ((a) = ((int)(a) < (int)(b) ? (a) : (b)))

#define XI_ICON_BUTTON_CLASS_NAME ("XiIconButton")
#define XI_EXT_LIST_CLASS_NAME ("XiExtendedList")
#define XI_ILIST_CLASS_NAME ("XiIList")
#define XI_EXT_18_LIST_CLASS_NAME ("XiExtended18List")
#define XI_I18LIST_CLASS_NAME ("XiI18List")

#if (XmVersion == 1002)
#define _XmDrawShadow(dpy, win, tgc, bgc, thick, x, y, w, h) \
    (_XmDrawShadows((dpy), (win), (tgc), (bgc), (x), (y), \
		    (w), (h), (thick), XmSHADOW_OUT))
#endif

/************************************************************
*	GLOBAL DECLARATIONS
*************************************************************/

#if defined(__cplusplus)
extern "C" {
#endif

    /*
     * Be sure to add new definitions for BX when adding new functions
     * so that we do no conflict with the versions of the widgets built
     * for BX.
     */

#ifdef _NO_PROTO
    extern XtGeometryResult _XiRequestNewSize(), _XiHWQuery();
    extern void _XiGetFocus(), _XiFilterArgs(), _XiSetValuesOnChildren(); 
    extern void _XiMoveWidget(), _XiResizeWidget(), _XiConfigureObject();
    extern Boolean _XiGadgetWarning();
    extern String _XiGetMBStringFromXmString();

    /*
     * Context Managment Routines.
     */
    
    extern Boolean _XiGetContextData(), _XiUtilIsSubclassByNameQ();
    extern void _XiSetContextData(), _XiDeleteContextData();
#else
    extern void _XiMoveWidget(Widget, Position, Position);
    extern void _XiResizeWidget(Widget, Dimension, Dimension, Dimension);
    extern void _XiConfigureWidget(Widget, Position, Position, 
			    Dimension, Dimension, Dimension);

    extern XtGeometryResult _XiRequestNewSize(Widget, Boolean, Dimension,
					      Dimension,
					      Dimension *, Dimension *);
    
    extern XtGeometryResult _XiHWQuery(Widget, XtWidgetGeometry*,
				       XtWidgetGeometry *);
    
    extern void _XiGetFocus(Widget, XEvent *, String *, Cardinal *);
    
    extern void _XiFilterArgs(ArgList, Cardinal, String *,
			      ArgList *, Cardinal *);

    extern void _XiSetValuesOnChildren(Widget, ArgList, Cardinal);
    
    extern Boolean _XiGadgetWarning(Widget);

    extern String _XiGetMBStringFromXmString(XmString);

    /*
     * Context Managment Routines.
     */
    
    extern void _XiSetContextData(Widget, XContext, XtPointer);
    extern void _XiDeleteContextData(Widget, XContext);
    extern Boolean _XiGetContextData(Widget, XContext, XtPointer *);
    extern Boolean _XiUtilIsSubclassByNameQ(Widget, XrmQuark);
#ifdef DEMO
    extern void _XiInitialIzeConverters(Widget);
#endif
#endif /* _NO_PROTO */

#if defined(__cplusplus)
}
#endif
/************************************************************
*	EXTERNAL DECLARATIONS
*************************************************************/

#if defined(VMS) || defined(__VMS)
#include <X11/apienvrst.h>
#endif

#endif /* __XiP_h__ */
