#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef VMS
#define strdup(s) strcpy(malloc(strlen(s)+1), (s))
#endif

typedef struct _RscRec {
    struct _RscRec *next;
    char	   *string;
    char           *name;
    int            index;
} RscRec, *Rsc;

static void
usage(cmd)
    char *cmd;
{
    int i, len = (cmd == NULL ? 0 : strlen(cmd)) + 8;

    fprintf(stderr,
	    "usage: %s [-c <c-file-name>] [-h <h-file-name] [-s <var-name>]\n",
	    cmd);
    for( i = 0; i < len; ++i ) fputc(' ', stderr);
    fprintf(stderr, "[-x <cpp-flag> <alt-var-name>]\n");
    exit(1);
}

int
main(argc, argv)
    int  argc;
    char *argv[];
{
    char *dot_c_name = "Unnamed.c",
         *dot_h_name = "Unnamed.h",
         *var_name   = "UnnamedStrings",
         *cpp_name   = NULL,
         *def_name   = NULL,
         *cmd = ((cmd = strrchr(argv[0], '/')) != NULL ? ++cmd : argv[0]),
         buf[512], *tmp, *tmp2, *tmp3;
    FILE *dot_c, *dot_h;
    int  i, idx, len;
    Rsc  rscs = NULL, next;

    for( i = 1; i < argc; ++i )
    {
	if( strcmp(argv[i], "-c") == 0 )
	{
	    if( ++i >= argc ) usage(cmd);
	    dot_c_name = argv[i];
	}
	else if ( strcmp(argv[i], "-h") == 0 )
	{
	    if( ++i >= argc ) usage(cmd);
	    dot_h_name = argv[i];
	}
	else if ( strcmp(argv[i], "-s") == 0 )
	{
	    if( ++i >= argc ) usage(cmd);
	    var_name = argv[i];
	}
	else if ( strcmp(argv[i], "-x") == 0 )
	{
	    if( ++i >= argc ) usage(cmd);
	    cpp_name = argv[i];
	    if( ++i >= argc ) usage(cmd);
	    def_name = argv[i];
	}
	else
	{
	    usage(cmd);
	}
    }

    while( !feof(stdin) && gets(buf) != NULL )
    {
	if( (next = (Rsc) malloc(sizeof(RscRec))) == NULL )
	{
	    fprintf(stderr, "%s: out of memory reading input\n",
		    cmd);
	    exit(1);
	}
	next->string = strdup(buf);
	tmp = strchr(next->string, 'N');
	tmp2 = strchr(next->string, 'C');
	tmp3 = strchr(next->string, 'R');
	if( tmp == NULL || (tmp2 != NULL && tmp2 < tmp) ) tmp = tmp2;
	if( tmp == NULL || (tmp3 != NULL && tmp3 < tmp) ) tmp = tmp3;

	if( tmp == NULL )
	{
	    next->name = NULL;
	}
	else
	{
	    next->name = ++tmp;
	}
	
	next->next = rscs;
	rscs = next;
    }

    if( (dot_h = fopen(dot_h_name, "w+")) == NULL )
    {
	sprintf(buf, "%s: unable to open output file \"%s\":",
		cmd, dot_h_name);
	perror(buf);
	exit(1);
    }

    if( (dot_c = fopen(dot_c_name, "w+")) == NULL )
    {
	fclose(dot_h);
	sprintf(buf, "%s: unable to open output file \"%s\":",
		cmd, dot_c_name);
	perror(buf);
	exit(1);
    }

    fprintf(dot_c, "/* This file is automatically generated. */\n");
    fprintf(dot_c, "/* Do not edit. */\n\n");
    fprintf(dot_c, "#ifdef __cplusplus\n");
    fprintf(dot_c, "extern \"C\" {\n");
    fprintf(dot_c, "#endif\n\n");
    fprintf(dot_c, "#if __STDC__\n");
    fprintf(dot_c, "#define Const const\n");
    fprintf(dot_c, "#else\n");
    fprintf(dot_c, "#define Const /**/\n");
    fprintf(dot_c, "#endif\n\n");
    if( cpp_name != NULL )
    {
	fprintf(dot_c, "#ifdef %s\n", cpp_name);
	fprintf(dot_c, "Const char %s[] = {\n", def_name);
	fprintf(dot_c, "#else\n");
    }
    fprintf(dot_c, "Const char %s[] = {\n", var_name);
    if( cpp_name != NULL )
    {
	fprintf(dot_c, "#endif\n");
    }
    for( idx = 0, next = rscs; next != NULL; next = next->next )
    {
	if( next->name == NULL ) continue;
	next->index = idx;
	len = strlen(next->name);
	idx += (len+1);
	for( i = 0; i < len; ++i )
	{
	    fprintf(dot_c, "\'%c\',", next->name[i]);
	}
	fprintf(dot_c, "0,\n");
    }
    fprintf(dot_c, "};\n\n");
    fprintf(dot_c, "#ifdef __cplusplus\n");
    fprintf(dot_c, "}\n");
    fprintf(dot_c, "#endif\n");
    fclose(dot_c);

    strcpy(buf, dot_h_name);
    if( (tmp = strchr(buf, '.')) != NULL ) *tmp = '\0';

    fprintf(dot_h, "/* This file is automatically generated. */\n");
    fprintf(dot_h, "/* Do not edit. */\n\n");
    fprintf(dot_h, "#ifndef __%s_h__\n", buf);
    fprintf(dot_h, "#define __%s_h__\n\n", buf);
    fprintf(dot_h, "#if defined(VMS) || defined (__VMS)\n");
    fprintf(dot_h, "#include <X11/apienvset.h>\n");
    fprintf(dot_h, "#endif\n\n");
    fprintf(dot_h, "#ifdef __cplusplus\n");
    fprintf(dot_h, "extern \"C\" {\n");
    fprintf(dot_h, "#endif\n\n");
    fprintf(dot_h, "#ifdef STRING_DEFINES\n");
    for( next = rscs; next != NULL; next = next->next )
    {
	if( next->name == NULL ) continue;

	fprintf(dot_h, "#ifndef %s\n", next->string);
	fprintf(dot_h, "#define %s \"%s\"\n", next->string, next->name);
	fprintf(dot_h, "#endif\n");
    }
    fprintf(dot_h, "#else\n");
    fprintf(dot_h, "#if __STDC__\n");
    fprintf(dot_h, "#define __Const__ const\n");
    fprintf(dot_h, "#else\n");
    fprintf(dot_h, "#define __Const__ /**/\n");
    fprintf(dot_h, "#endif\n");
    if( cpp_name != NULL )
    {
	fprintf(dot_h, "#ifdef %s\n", cpp_name);
	fprintf(dot_h, "extern __Const__ char %s[];\n", def_name);
	fprintf(dot_h, "#define %s %s\n", var_name, def_name);
	fprintf(dot_h, "#else\n");
    }
    fprintf(dot_h, "extern __Const__ char %s[];\n", var_name);
    if( cpp_name != NULL )
    {
	fprintf(dot_h, "#endif\n");
    }
    fprintf(dot_h, "#undef __Const__\n\n");
    for( next = rscs; next != NULL; next = next->next )
    {
	if( next->name == NULL ) continue;
	fprintf(dot_h, "#ifndef %s\n", next->string);
	fprintf(dot_h, "#define %s ((char*)&%s[%d])\n", next->string,
		var_name, next->index);
	fprintf(dot_h, "#endif\n");
    }
    fprintf(dot_h, "#endif\n");
    fprintf(dot_h, "#ifdef __cplusplus\n");
    fprintf(dot_h, "}\n");
    fprintf(dot_h, "#endif\n\n");
    fprintf(dot_h, "#if defined(VMS) || defined (__VMS)\n");
    fprintf(dot_h, "#include <X11/apienvrst.h>\n");
    fprintf(dot_h, "#endif\n\n");
    fprintf(dot_h, "#endif\n");    
    fclose(dot_h);

    return( 0 );
}


