// Revision: 16 1.4.1.2 source/multmed/immdev1.cpp, multimedia, ioc.v400, 000920 
/*******************************************************************************
* FILE NAME: immdev1.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in immdevnh.hpp and some in immdev.hpp.                                    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#pragma priority( -2147481324 )

#define INCL_WINSYS

extern "C" {
  #define IC_MULTIMEDIA
  #include <immdefs.h>
}

#include <iwindow.hpp>
#include <immowin.hpp>
#include <immdev.hpp>
#include <immdevnh.hpp>
#include <ievent.hpp>
#include <inotifev.hpp>
#include <iexcept.hpp>
#include <icconst.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMMDEV1_CPP_
  #include <ipagetun.h>
#endif

// Initialize static objects and notification strings.
INotificationId const IMMDevice::commandNotifyId           = "IMMDevice::commandNotify";
INotificationId const IMMDevice::passDeviceId              = "IMMDevice::passDevice";
INotificationId const IMMDevice::cuePointId                = "IMMDevice::cuePoint";
INotificationId const IMMDevice::positionChangeId          = "IMMDevice::positionChange";
INotificationId const IMMDevice::deviceEventId             = "IMMDevice::deviceEvent";

/*------------------------------------------------------------------------------
| IMMDevice::enableNotification                                                |
|                                                                              |
| Flags this part as being ready to accept notifications.  A Notifier, like    |
| IMMDevice, does not send notifications to observer objects until it          |
| is enabled.                                                                  |
|                                                                              |
| In this class, we override the INotifier function to allocate the correct    |
| notification handler. We use our msg window to manage the handler & boolean. |
------------------------------------------------------------------------------*/
IMMDevice& IMMDevice::enableNotification (bool enabled)
{
  if (!notificationHandler())
  {
    setNotificationHandler( new IMMDeviceNotifyHandler() );
  }
  deviceWindow().enableNotification(enabled);    // Enable our proxy window, so our notification
                                                 // handler gets called (also enables IWindow notifIds e.g., destroyId)
  Inherited::enableNotification(enabled);                   // Enable this device also, so notifyObservers works
  return *this;
}


/*------------------------------------------------------------------------------
| IMMDevice::notificationHandler                **aj**                         |
|                                                                              |
| Returns a pointer to the notification handler                                |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler* IMMDevice::notificationHandler    ( ) const
{
  IMMDeviceNotifyHandler* notifyHandler =
           (IMMDeviceNotifyHandler*)(deviceWindow().notificationHandler());
  return notifyHandler;
}

/*------------------------------------------------------------------------------
| IMMDevice::setNotificationHandler           **aj**                           |
|                                                                              |
| Sets the notification handler to the passed value.                           |
------------------------------------------------------------------------------*/
IMMDevice& IMMDevice::setNotificationHandler (
                                     IMMDeviceNotifyHandler* notifyHandler )
{
  IASSERTSTATE(notifyHandler!=0);
  deviceWindow().setNotificationHandler(notifyHandler);
  return *this;
}


/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::IMMDeviceNotifyHandler                               |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler::IMMDeviceNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::~IMMDeviceNotifyHandler                              |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IMMDeviceNotifyHandler::~IMMDeviceNotifyHandler()
{}

/*------------------------------------------------------------------------------
| IMMDeviceNotifyHandler::dispatchHandlerEvent                                 |
|                                                                              |
| This notifies IMMDevice Windows of the following events:                     |
|                                                                              |
| mediaLoaded(Id) - either loaded (true) or unloaded (false)                   |
------------------------------------------------------------------------------*/

bool IMMDeviceNotifyHandler::dispatchHandlerEvent (IEvent& anEvent)
{
   IMODTRACE_DEVELOP("IMMDeviceNotifyHandler::dispatchHandlerEvent") ;

  switch (anEvent.eventId())
  {
#ifdef IC_WIN
     case IC_UM_MM_MCI_CMDNOTIFY:    // in Win32, MM_MCI_NOTIFY is missing info
                                     // so we work around it with this event
#endif
#ifdef IC_PM
    case MM_MCINOTIFY:
#endif
      {
         ITRACE_DEVELOP("IMMDeviceNotifyHandler::mm_mcinotify") ;
      IMMNotifyEvent notifyEvent(anEvent);
      IMMDevice* device = notifyEvent.device();
      device->notifyObservers(
                     INotificationEvent(IMMDevice::commandNotifyId,
                                        *device,true, &notifyEvent));
      break;
      }

    case MM_MCIPASSDEVICE:
      {
      IMMPassDeviceEvent passEvent(anEvent);
      IMMDevice* device = passEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::passDeviceId,
                                     *device,true, &passEvent));
      break;
      }

#ifdef IC_WIN
    case IC_UM_MM_MCI_POSITIONCHANGE:
#endif
#ifdef IC_PM
    case MM_MCIPOSITIONCHANGE:
#endif
      {
         ITRACE_DEVELOP("IMMDeviceNotifyHandler::mm_mcipositionchange") ;
      IMMPositionChangeEvent posEvent(anEvent);
      IMMDevice* device = posEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::positionChangeId,
                                     *device,true, &posEvent));
      break;
      }

#ifdef IC_PM
    case MM_MCICUEPOINT:
      {
      IMMCuePointEvent cueEvent(anEvent);
      IMMDevice* device = cueEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::cuePointId,
                                     *device,true, &cueEvent));
      break;
      }

    case MM_MCIEVENT:
      {
      IMMDeviceEvent devEvent(anEvent);
      IMMDevice* device = devEvent.device();
      device->notifyObservers(
                  INotificationEvent(IMMDevice::deviceEventId,
                                     *device,true, &devEvent));
      break;
      }
#endif
     default:
        ITRACE_DEVELOP("IMMDeviceNotifyHandler::default"+ IString(anEvent.eventId())) ;
         break;
  }  /* end switch */

   return (Inherited::dispatchHandlerEvent(anEvent));
}

