/*******************************************************************************
* FILE NAME: igelipse.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igelipse.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Revision: 53 1.3.1.3 source/ographic/igelipse.cpp, graphics, ioc.v400, 000920 

#include <ibase.hpp>

#if ( IC_OBSOLETE <= IC_OBSOLETE_3 )
#ifdef IC_PMWIN

#include <igrafctx.hpp>
#include <igelipse.hpp>
#include <igbundle.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGELIPSE_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGEllipse::IGEllipse                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse::IGEllipse( const IGEllipse& ellipse)
  : IGEllipse::Inherited(ellipse), fRect(ellipse.fRect)
{
}

/*------------------------------------------------------------------------------
| IGEllipse::IGEllipse                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse::IGEllipse( const IRectangle& rectangle )
  : IGEllipse::Inherited(), fRect(rectangle)
{
}

/*------------------------------------------------------------------------------
| IGEllipse::IGEllipse                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse::IGEllipse( const IPoint& point, unsigned long Radius )
  : IGEllipse::Inherited(), fRect( IPoint(point.x()-Radius, point.y()-Radius),
                                   IPoint(point.x()+Radius, point.y()+Radius))
{
}

/*------------------------------------------------------------------------------
| IGEllipse::IGEllipse                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse::IGEllipse( const IPoint& point, unsigned long xRadius,
                      unsigned long yRadius )
  : IGEllipse::Inherited(), fRect( IPoint(point.x()-xRadius, point.y()-yRadius),
                                   IPoint(point.x()+xRadius, point.y()+yRadius))
{
}

/*------------------------------------------------------------------------------
| IGEllipse::~IGEllipse                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse::~IGEllipse()
{
}

/*------------------------------------------------------------------------------
| IGEllipse::drawOn                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse& IGEllipse::drawOn( IGraphicContext& gc )
{
  gc.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGEllipse::setEnclosingRect                                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse& IGEllipse::setEnclosingRect( const IRectangle& rectangle )
{
  fRect = rectangle;
  return *this;
}

/*------------------------------------------------------------------------------
| IGEllipse::enclosingRect                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IRectangle IGEllipse::enclosingRect() const
{
  return fRect;
}

/*------------------------------------------------------------------------------
| IGEllipse::operator=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGEllipse& IGEllipse::operator= ( const IGEllipse& ellipse )
{
  if ( this == &ellipse )
    return *this;

  this->Inherited::operator=(ellipse);
  fRect = ellipse.fRect;
  return *this;
}

/*------------------------------------------------------------------------------
| IGEllipse::operator==                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
bool IGEllipse::operator==( const IGEllipse& ellipse ) const
{
  return (fRect == ellipse.fRect && this->Inherited::operator==(ellipse));
}

/*------------------------------------------------------------------------------
| IGEllipse::operator!=                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
bool IGEllipse::operator!=( const IGEllipse& ellipse ) const
{
  return !( *this == ellipse );
}

#endif // IC_PMWIN
#endif // IC_OBSOLETE
