/*******************************************************************************
* FILE NAME: iuiobs.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of functions declared in the UI      *
*   Library that are still dependent upon old graphics.                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Revision: 75 1.7.1.6 source/ographic/iuiobs.cpp, graphics, ioc.v400, 000920 

extern "C" {
  #define INCL_WINBUTTONS
  #include <iwindefs.h>
}

#include <ibase.hpp>

#if ( IC_OBSOLETE <= IC_OBSOLETE_3 )
#ifdef IC_PMWIN

#include <itrace.hpp>
#include <igrafctx.hpp>
#include <icoordsy.hpp>
#include <iglist.hpp>


#include <ibidiset.hpp>
#include <ipoint.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
   #define _IUIOBS_CPP_
   #include <ipagetun.h>
#endif

IBidiSettings::IBidiSettings ( const IGraphicContext& graphicContext )
              :fBidiSettingsData(0)
{
  initialize(graphicContext.handle());
}

IBidiSettings& IBidiSettings::apply ( const IGraphicContext& graphicContext ) const
{
  apply ( graphicContext.handle() );
  return (IBidiSettings&)*this;
}


#include <ipainevt.hpp>
#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IPaintEvent::setGraphicContext                                               |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::setGraphicContext ( const IGraphicContext& gc,
                                              bool setClipRegion )
{
  return this->setGraphicContext( gc.handle(), setClipRegion );
}
#endif


#include <icustbev.hpp>
#include <icustbut.hpp>
/*------------------------------------------------------------------------------
| ICustomButtonDrawEvent::graphicContext                                       |
------------------------------------------------------------------------------*/
IGraphicContext& ICustomButtonDrawEvent :: graphicContext () const
{
  ICustomButtonDrawEvent* localthis = (ICustomButtonDrawEvent*)this;
  if (!fGraphicContext)
  {
#ifndef IC_WIN
     localthis->fGraphicContext = new IGraphicContext(
        controlWindow()->handle()) ;
#else
     localthis->fGraphicContext = new IGraphicContext(
        IPresSpaceHandle( ((LPDRAWITEMSTRUCT)this->parameter2().asUnsignedLong())->hDC,
                          this->controlHandle() ) );
#endif
  }

  return *((IGraphicContext*)(localthis->fGraphicContext));
}



#include <idrawcv.hpp>


ISize oldCalcMinimumSize ( const IDrawingCanvas* drawingCanvas, const ISize& inheritedMinSize)
{
  IGraphicContext *gc(drawingCanvas->graphicContext());
  bool createdGC(false);

  IRectangle listBoundingRect;
  ISize      minSize;

  if (!gc)
  {
    gc = new IGraphicContext( drawingCanvas->handle() );
    createdGC = true;
  }

  if (drawingCanvas->graphicList() && !drawingCanvas->graphicList()->isEmpty())
  {
    listBoundingRect = drawingCanvas->graphicList()->boundingRect( *gc );
    minSize.setWidth( listBoundingRect.maxX() );
    minSize.setHeight( listBoundingRect.maxY() );
  }

  if (createdGC)
  {
    delete gc;
    gc = 0;
  }

  minSize = minSize.maximum( inheritedMinSize );

  return minSize;
}

bool oldPaintWindow( IDrawingCanvas* drawingCanvas, IPaintEvent& paintEvent  )
{
  bool          didDraw(false);
  bool          createdGC(false);
  IGraphicContext* gc(drawingCanvas->graphicContext());

  if ( gc )
  {
    paintEvent.setGraphicContext( *gc );
  }
  else
  {
    if (paintEvent.presSpaceHandle())
    {
      gc = new IGraphicContext(paintEvent.presSpaceHandle());
      createdGC = true;
    }
  }

  if (paintEvent.rect().area())
  {
    paintEvent.clearBackground(drawingCanvas->backgroundColor());
    if (drawingCanvas->graphicList())
    {
      if (ICoordinateSystem::isConversionNeeded())
        gc->setRecoordinationHeight( drawingCanvas->size().height() );
      drawingCanvas->graphicList()->drawOn(*gc);
    }

    if (createdGC)
    {
      delete gc;
      gc = 0;
    }
    didDraw = true;
  }
  return didDraw;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::graphicList                                                  |
|                                                                              |
------------------------------------------------------------------------------*/
IGList* IDrawingCanvas::graphicList( ) const
{
  return fList;
}

static bool gAddressesSet = false;
/*------------------------------------------------------------------------------
| IDrawingCanvas::setGraphicList                                               |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas& IDrawingCanvas::setGraphicList( IGList* list )
{
  if(gAddressesSet == false)
  {
     setOldCalcMinimumSize(&oldCalcMinimumSize);
     setOldPaintWindow(&oldPaintWindow);
     gAddressesSet = true;
  }
  fList = list;
  fListType = kOldGraphics;
  setLayoutDistorted(IWindow::layoutChanged, 0);
  return *this;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::setGraphicContext                                            |
|                                                                              |
------------------------------------------------------------------------------*/
IDrawingCanvas& IDrawingCanvas::setGraphicContext( IGraphicContext* pGC )
{
  if(gAddressesSet == false)
  {
     setOldCalcMinimumSize(&oldCalcMinimumSize);
     setOldPaintWindow(&oldPaintWindow);
     gAddressesSet = true;
  }
  fGc = pGC;
  fContextType = kOldGraphics;
  return *this;
}

/*------------------------------------------------------------------------------
| IDrawingCanvas::graphicContext                                               |
|                                                                              |
------------------------------------------------------------------------------*/
IGraphicContext* IDrawingCanvas::graphicContext( ) const
{
  return fGc;
}

#endif // IC_PMWIN
#endif // ( IC_OBSOLETE <= IC_OBSOLETE_3 )
